/*
 *  Rotator.cpp - 	Rotator effect
 *               	Written by Tinic Urou
 *					You can manipulate speed with '+' and '-'
 *					Motion Blur with 'b'
 *               	Public domain
 */

#include <AppKit.h>
#include <InterfaceKit.h>
#include <KernelKit.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

// Types
typedef unsigned char 	UBYTE;
typedef unsigned short 	UWORD;
typedef unsigned long 	ULONG;
typedef char 			BYTE;
typedef short 			WORD;
typedef long 			LONG;

// Constants

const 	ULONG 			APP_SIGNATURE = '????';
const 	LONG 			DISPLAY_X =320;
const 	LONG 			DISPLAY_Y =240;

		UWORD			SPEED=12;
		bool			BLUR=FALSE;

class 					RotatorWindow;
class					KeyView;


extern 	UBYTE			pic[];
		WORD			FastSin	[1024];
		WORD			Fast	[1024];
		UWORD 			AngleMulti[320*240*2];
		ULONG			xmod;
		UWORD 			xo=0;
		UBYTE			*bits;
		color_map		*sys_col;

// Window/Application messages

const 	LONG			MSG_REDRAW = 1;

// Application object
class Rotator : public BApplication 
{
	public:
						Rotator();
			void 		ReadyToRun(void);
			bool 		QuitRequested(void);
			void 		AboutRequested(void);

	private:
	
			RotatorWindow 	*the_window;
			BBitmap 	*the_bitmap;
};

// Window object
class RotatorWindow : public BWindow 
{
	public:
	
						RotatorWindow(int width, int height, BBitmap *bitmap);
			bool 		QuitRequested(void);
			void 		MessageReceived(BMessage *msg);

	private:
			void		recalc(void);
	static	void 		draw_thread(int line);

			BBitmap 	*the_bitmap;
			BView		*main_view;	
			KeyView		*key_view;
			thread_id	t1,t2;	
};

// View object for keydown
class KeyView : public BView
{
	public:

	KeyView():BView(BRect(0,0,1,1),"",NULL,NULL)
	{
	};
	
	void KeyDown(ulong aKey)
	{
		switch(aKey)
		{
			case	'-':
					SPEED--;
					break;
			case	'+':
					SPEED++;
					break;
			case	'b':
			case	'B':
					if(BLUR)BLUR=FALSE;
					else BLUR=TRUE;
					break;
		}
	};
};

// Create application object and start it
int main(int argc, char **argv)
{	
	Rotator *the_app;

	srand(system_time());

	the_app = new Rotator();
	the_app->Run();
	delete the_app;

	return 0;
}


/*
 *  Application constructor
 */

Rotator::Rotator() : BApplication(APP_SIGNATURE)
{
	the_window = NULL;
	the_bitmap = NULL;
}

/*
 *	Thats from me!
 */

void Rotator::AboutRequested(void) 
{
	char str[256];
	sprintf(str, "Rotator by Tinic Urou\n<5uro@informatik.uni-hamburg.de>\nFreely distributable.");
	BAlert *the_alert = new BAlert("", str, "OK");
	the_alert->Go();
}


/*
 *  Open window and start rendering
 */

void Rotator::ReadyToRun(void)
{
	// Allocate the bitmap		
	the_bitmap = new BBitmap(BRect(0, 0, DISPLAY_X-1, DISPLAY_Y-1), B_COLOR_8_BIT);

	bits = (UBYTE *)the_bitmap->Bits();

	// Open window
	the_window = new RotatorWindow(DISPLAY_X, DISPLAY_Y, the_bitmap);

	// Start action
	the_window->PostMessage(MSG_REDRAW);
}

/*
 *  Quit requested (either by menu or by closing the window)
 */

bool Rotator::QuitRequested(void)
{
	// Make sure that the window closes first
	if (BApplication::QuitRequested()) 
	{
		if (the_bitmap)	delete the_bitmap;
		return TRUE;
	}
	return FALSE;
}


/*
 *  Window constructor
 */

RotatorWindow::RotatorWindow(int width, int height, BBitmap *bitmap) : BWindow(BRect(0, 0, width-1, height-1), "Rotator by Tinic", B_TITLED_WINDOW, B_NOT_RESIZABLE | B_NOT_ZOOMABLE)
{
	the_bitmap = bitmap;
	
	xmod = bitmap->BytesPerRow();
	
	WORD	*f=FastSin;
	for (ULONG x=0; x<512; x++)
	{
		*f++=(( sin((x*(M_PI*2))/512))*(256));
		*f++=((-cos((x*(M_PI*2))/512))*(256));
	}
	
	UWORD 	*p=AngleMulti;	
	for (LONG y=0;y<DISPLAY_Y;y++)
	{	
		for (LONG x=0;x<DISPLAY_X;x++)
		{	
			*p++=sqrt((x-DISPLAY_X/2)*(x-DISPLAY_X/2)+(y-DISPLAY_Y/2)*(y-DISPLAY_Y/2));
			*p++=atan2(x-DISPLAY_X/2,y-DISPLAY_Y/2)*512/M_PI;
		}
	}
	
	sys_col=system_colors();

	// Speed up drawing
	
	SetDiscipline(FALSE);

	// Move window to right position
	MoveTo(80, 60);

	// Create bitmap view
	Lock();
	key_view	= new KeyView();
	AddChild(key_view);
	key_view->MakeFocus();
	main_view 	= new BView(BRect(0, 0, DISPLAY_X-1, DISPLAY_Y-1),"", B_FOLLOW_NONE,NULL);
	AddChild(main_view);
	Unlock();

	// Show the window
	Show();
}


/*
 *  Closing the window quits the program
 */

bool RotatorWindow::QuitRequested(void)
{
	be_app->PostMessage(B_QUIT_REQUESTED);
	return TRUE;
}


/*
 *  Handles redraw messages
 */

void RotatorWindow::MessageReceived(BMessage *msg)
{	
	switch (msg->what) 
	{
		case MSG_REDRAW:  // Redraw bitmap
			Lock();
			main_view->DrawBitmapAsync(the_bitmap);
			recalc();
			main_view->Sync();
			PostMessage(MSG_REDRAW);
			Unlock();
			break;
		default:
			BWindow::MessageReceived(msg);
	}
}

/*
 *  Main calculation
 */

void RotatorWindow::draw_thread(int line)
{
	ULONG 		mod=xmod;
	UWORD 		*todo=AngleMulti+(line*DISPLAY_X*2);
	UBYTE 		*q=bits+(line*mod);
	UWORD 		off=xo;
	ULONG		buf;
	
	if(!BLUR)
	{	
		register UBYTE *p;
		for(int y=0;y<DISPLAY_Y/2;y++)
		{	
			p=q;q+=mod;
			for(int x=0;x<(DISPLAY_X/4);x++)
			{	
				// Use as much registers as possible ;-)
				// We nEEEEEd L2 Cache!!!
				
				UWORD	bar1=*todo++;
				WORD	*foo1=((off+*todo++)&0x3FE)+Fast;
				UWORD	bar2=*todo++;
				WORD	*foo2=((off+*todo++)&0x3FE)+Fast;
				UWORD	bar3=*todo++;
				WORD	*foo3=((off+*todo++)&0x3FE)+Fast;
				UWORD	bar4=*todo++;
				WORD	*foo4=((off+*todo++)&0x3FE)+Fast;
								
				UWORD	y1=(((*foo1++)*bar1)    +(DISPLAY_X*128))&0xFF00;
				UBYTE	x1=(((*foo1  )*bar1)>>8)+(DISPLAY_Y/  2);
				UWORD	y2=(((*foo2++)*bar2)    +(DISPLAY_X*128))&0xFF00;
				UBYTE	x2=(((*foo2  )*bar2)>>8)+(DISPLAY_Y/  2); 
				UWORD	y3=(((*foo3++)*bar3)    +(DISPLAY_X*128))&0xFF00;
				UBYTE	x3=(((*foo3  )*bar3)>>8)+(DISPLAY_Y/  2);
				UWORD	y4=(((*foo4++)*bar4)    +(DISPLAY_X*128))&0xFF00;
				UBYTE	x4=(((*foo4  )*bar4)>>8)+(DISPLAY_Y/  2);
				
				*p++=pic[y1+x1];
				*p++=pic[y2+x2];
				*p++=pic[y3+x3];
				*p++=pic[y4+x4];		
			}
		}
	}
	else
	{
		rgb_color *col_list=sys_col->color_list;
		uchar *index=sys_col->index_map;
		UBYTE *p;

		for (int y=0;y<DISPLAY_Y/2;y++)
		{	
			p=q;q+=mod;
			for (int x=0;x<DISPLAY_X;x++)
			{			
				UWORD	bar=*todo++;
				WORD	*foo=((off+*todo++)&0x3FE)+Fast;
				UWORD	y1=(((*foo++)*bar)    +(DISPLAY_X*128))&0xFF00;
				UBYTE	x1=(((*foo  )*bar)>>8)+(DISPLAY_Y/  2) &0xFF;

				rgb_color	oldcolor,newcolor;
			
				oldcolor=col_list[*p];
				newcolor=col_list[pic[y1+x1]];
			
				*p++=index
				[
					(((oldcolor.red		+newcolor.red	)&0x1F0)<<6)|
					(((oldcolor.green	+newcolor.green	)&0x1F0)<<1)|
					(((oldcolor.blue	+newcolor.blue	)&0x1F0)>>4)
				]; 
			}
		}
	}
}

void RotatorWindow::recalc(void)
{
	static	UWORD m=0;

	WORD	*f=Fast,*g=FastSin;
	LONG	returnvalue;
	WORD	tmp;
			
	for (ULONG x=0; x<512; x++)
	{
		tmp=((FastSin[m]+292)>>1);
		
		*f++=((*g++)*tmp)/100;
		*f++=((*g++)*tmp)/100;
	}
		
	resume_thread(t1=spawn_thread((thread_entry)draw_thread,"Rotator Draw Thread 1",B_NORMAL_PRIORITY,(void *)0));
	resume_thread(t2=spawn_thread((thread_entry)draw_thread,"Rotator Draw Thread 2",B_NORMAL_PRIORITY,(void *)(DISPLAY_Y/2)));
	wait_for_thread(t1,&returnvalue);
	wait_for_thread(t2,&returnvalue);		

	xo+=SPEED;
	m +=SPEED+2;m&=0x3FE;
}
